/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  int ADopenPlay (AudDev, Flags)
  void ADclosePlay (void)

Purpose:
  Open a Sun audio device for playback
  Close a Sun audio device

Description:
  This routine opens or closes an audio device.  The open function opens an
  audio device for playback.  It also sets up a trap for SIGINT which flushes
  the output and closes the device.  The close function drains the output and
  restores the port and volume settings of the audio device.

Parameters:
  <-  int ADopenPlay
      File descriptor for the audio device
   -> const char AudDev[]
      String containing the audio device name or the name of an environment
      variable (with a leading $)
   -> int Flags
      Control variable, the OR of the following values
       1 - Wait flag.  If not set, an error is generated if the device is
           busy.  If set, wait for the device to become not busy.
       2 - Print flag.  Print the device name.

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.20 $  $Date: 1999/06/14 16:17:48 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: ADopenPlay.c 1.20 1999/06/14 AFsp-v5r0 $";

#include <errno.h>
#include <fcntl.h>		/* open */
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>		/* EXIT_FAILURE */
#include <stropts.h>		/* stream ioctl commands */
#include <unistd.h>		/* close, ioctl */

#include <libtsp.h>
#include <AD.h>

static ADEV Adev;

#define FSET(flag,mask)		(((flag) & (mask)) != 0)
#define PGM		(UTgetProg ())

static void
ADsigintP (int sig);	/* prototype for the signal handler */


int
ADopenPlay (const char AudioDev[], int Flags)

{
  const char *audiodev;
  int iost;
  int audio_fd;
  audio_info_t aif;
  void (*Phandler)();

#ifdef SOLARIS
  int io_flags;
  audio_device_t AD;
#else
  int AD;
#endif

/* Check for an environment variable */
  if (AudioDev[0] == '$') {
    audiodev = getenv (&AudioDev[1]);
    if (audiodev == NULL)
      audiodev = AUDIO_DEV_DEFAULT;
  }
  else
    audiodev = AudioDev;

/* Open the audio device */
  audio_fd = open (audiodev, O_WRONLY | O_NDELAY);
  if (audio_fd < 0 && errno == EBUSY) {
    if (FSET (Flags, AD_WAIT_FLAG))
      UTwarn ("%s: %s \"%s\"", PGM, ADM_WaitDev, audiodev);
    else
      UTerror ("%s: %s", PGM, ADM_DevBusy);

    audio_fd = open (audiodev, O_WRONLY);
    if (audio_fd >=0)
      UTwarn ("%s - %s", PGM, ADM_DevAtt);
  }
  if (audio_fd < 0)
    UTerror ("%s: %s \"%s\"", PGM, ADM_OpenErr, audiodev);

/* Set the audio device to blocking (turn off the O_NDELAY flag)
   - Solaris: undo the O_NDELAY flag (now in blocking mode).  This is necessary
     in order to play long files (avoids an "Invalid argument" error when
     writing to the device)
   - SunOS 4: O_NDELAY according to fcntl(5) documentation only applies to
     reads
*/
#ifdef SOLARIS
  io_flags = fcntl (audio_fd, F_GETFL);
  if (io_flags == -1)
    UTerror ("%s: %s \"%s\"", PGM, ADM_AccessErr, audiodev);
  if (fcntl (audio_fd, F_SETFL, io_flags & ~O_NDELAY) == -1)
    UTerror ("%s: %s \"%s\"", PGM, ADM_AccessErr, audiodev);
#endif

/* Get the device type */
  if (FSET (Flags, AD_PRINT_FLAG)) {
    iost = ioctl (audio_fd, AUDIO_GETDEV, &AD);
    if (iost == -1)
      UTerror ("%s: %s", PGM, ADM_GetErr);

#ifdef SOLARIS
    printf (ADMF_ADev, audiodev, MAX_AUDIO_DEV_LEN, AD.name);
#else
    if (AD < 0)
      AD = 0;
    if (AD >= NAD_type)
      AD = NAD_type - 1;
    printf (ADMF_ADevS, audiodev, AD_type[AD]);
#endif
  }

/* Get the device information */
  iost = ioctl (audio_fd, AUDIO_GETINFO, &aif);
  if (iost == -1)
    UTerror ("%s: %s", PGM, ADM_GetErr);
  if (aif.output_muted != 0)
    UTwarn ("%s: %s", PGM, ADM_Muted);

/* Save the volume, balance, port configuration */
  Adev.play.gain = aif.play.gain;
  Adev.play.port = aif.play.port;
  Adev.play.balance = aif.play.balance;
  Adev.output_muted = aif.output_muted;
  Adev.fd = audio_fd;

/* Set up a signal handler for SIGINT, restores the audio device settings */
  Phandler = signal (SIGINT, &ADsigintP);
  if (Phandler == SIG_ERR)
    UTerror ("%s, %s", PGM, ADM_SigHandler);

  return audio_fd;
}

/* Reset parameters and close the device */

void
ADclosePlay (void)

{
  audio_info_t aif;
  int iost;

/* Ensure output is done before resetting parameters */
  ADdrain (Adev.fd);

/* Reset volume, port and balance configuration */
  AUDIO_INITINFO (&aif);		/* Initialize the structure */
  aif.play.gain = Adev.play.gain;
  aif.play.port = Adev.play.port;
  aif.play.balance = Adev.play.balance;
  aif.output_muted = Adev.output_muted;
  iost = ioctl (Adev.fd, AUDIO_SETINFO, &aif);
  if (iost == -1)
    UTerror ("%s: %s", PGM, ADM_SetErr);

  close (Adev.fd);

}

/* Signal handler to close the audio device */

static void
ADsigintP (int sig)

{
  int iost;

  if (Adev.fd >= 0) {
    iost = ioctl (Adev.fd, I_FLUSH, FLUSHW);
    if (iost == -1)
      UTerror ("%s, %s", PGM, ADM_FlushErr);
    ADclosePlay ();
  }

  exit (EXIT_FAILURE);
}
