/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  void ADplay (int audio_fd, AFILE *AFp, long int Lim[2])

Purpose:
  Play an audio file

Description:
  This routine plays an audio file.  Data is read from the audio file, with
  output to an intermediate float buffer.  This data buffer is then converted
  to the appropriate data type for output to the audio device.  This routine
  waits for the output to finish before returning.

Parameters:
   -> int audio_fd
      File descriptor for the audio device
   -> AFILE *AFp
      Audio file pointer for the audio file
   -> const long int Lim[2]
      Play frame limits

Author / revision:
  P. Kabal  Copyright (C) 1998
  $Revision: 1.12 $  $Date: 1998/10/07 19:57:24 $

-------------------------------------------------------------------------*/

static char rcsid[] = "$Id: ADplay.c 1.12 1998/10/07 AFsp-v5r0 $";

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <unistd.h>		/* write */

#include <libtsp.h>
#include <libtsp/AFpar.h>
#include <libtsp/UTtypes.h>
#include <AD.h>

#define MINV(a, b)	(((a) < (b)) ? (a) : (b))
#define MAXV(a, b)	(((a) > (b)) ? (a) : (b))
#define ICEILV(n, m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define RUPV(n, m)	((m) * ICEILV ((n), (m)))

#define PGM		(UTgetProg ())

#define BUF_MIN_SAMP	20	/* Minimum number of samples to play */
#define BUF_TIME	0.83	/* Buffer size in seconds */
#define BUF_MSIZE	8192	/* Desirable size */
/* The AF read routines (except for text data) read from the file in chunks of
   8192 bytes.  Here we try to make the buffer size a multiple of 8192 samples.
   The buffer size in seconds is chosen so that for both 8000 Hz and 9600 Hz,
   we will use 8192 samples.
*/


void
ADplay (audio_fd, AFp, Lim)

     int audio_fd;
     AFILE *AFp;
     const long int Lim[2];

{
  float *Fbuf;
  long int offs, Nframe, Nrem, NchanI, NchanO, Novld;
  int Nr, Nval, Nbsamp, Nbytes, Nb;
  int eof, stereo_to_mono, i, k;
  float Sfreq;
  int Format;

/* Get the audio device output settings */
  ADgetOFormat (audio_fd, &Sfreq, &Format, &NchanO);

/* Check the number of input and output channels */
/*
  Valid combinations
    NchanI == NchanO
    NchanI == 2 && NchanO == 1
*/
  NchanI = AFp->Nchan;
  stereo_to_mono = (NchanI == 2 && NchanO == 1);
  assert (NchanI == NchanO || stereo_to_mono);
  eof = (Lim[1] == AD_LIM_UNDEF);

/* Choose the buffer size */
/*
  We pass a buffer to the audio device.  The write to the audio device
  returns as soon as this buffer has been copied into the audio device buffer.
  This means that we may pass through the file read / device write loop
  several times before the audio device buffer is full and the write blocks.

  The size of the buffer is determined by several factors.  Clearly for
  realtime playback, the read/conversion time for a buffer of data must be
  less than the playback time.
    - A large buffer averages out read/conversion time variations for
      multi-user systems with fragmented disk files.
    - The read/conversion time for a large buffer may delay the start of
      playback.
  Experience with the slowest read/conversion times (for text data files)
  indicates that even at 8 kHz sampling, the minimum buffer size for playback
  without glitches corresponds to about 1 sec.  But for text data files, the
  delay before playback is noticeable with such a buffer.

  The float buffer is overwritten by the output data (assumes that the output
  data size is no larger than a float)

  Odd behaviour (at least for Sparc 4 audioamd, SunOS 4.1.4):
  If the number of samples to be played is 12 or less, a subsequent drain
  command hangs forever.  Here we force the minimum transfer to be 20 samples.
*/
  Nbsamp = NchanI * RUPV ((int) (BUF_TIME * Sfreq), BUF_MSIZE); /* round up */
  assert (Nbsamp > 0);

  /* Extra long buffer (2 sec) for the slow-to-read text files */
  if (AFp->Format == FD_TEXT)
    Nbsamp *= 2;

  /* Number of frames to play */
  if (! eof) {
    Nframe = Lim[1] - Lim[0] + 1;

    /* Print the sample range */
    if (NchanI == 1)
      printf (ADMF_Samples, Lim[0], Lim[1], Nframe, Nframe / Sfreq);
    else
      printf (ADMF_Frames, Lim[0], Lim[1], Nframe, Nframe / Sfreq);

    Nrem = (int) MAXV (NchanI * Nframe, BUF_MIN_SAMP);
    Nbsamp = MINV (Nbsamp, Nrem);

  }
  else
    Nrem = LONG_MAX;
  
  Fbuf = (float *) UTmalloc (Nbsamp * sizeof (float));

/* ==================== Main loop */
  Novld = 0;
  offs = NchanI * Lim[0];

  while (Nrem > 0) {

    /* Read the data (with conversion) */
    Nval = (int) MINV (Nbsamp, Nrem);
    Nr = AFreadData (AFp, offs, Fbuf, Nval);

    if (Nr < Nval && eof) {
      Nval = Nr;
      Nrem = Nval;
    }
    offs += Nval;
    Nrem -= Nval;

    /* Create a mono signal from a stereo signal */
    if (stereo_to_mono) {
      Nval = Nval / 2;
      for (i = 0, k = 0; i < Nval; ++i, k += 2)
	Fbuf[i] = 0.5  * (Fbuf[k] + Fbuf[k+1]);
    }

    /* Convert to the output format */
    if (Format == FD_INT16) {
      Novld += ADfI2 (Fbuf, (int2_t *) Fbuf, Nval);
      Nbytes = 2 * Nval;
    }
    else if (Format == FD_MULAW8) {
      Novld += ADfMulaw (Fbuf, (uint1_t *) Fbuf, Nval);
      Nbytes = Nval;
    }
    else
      assert (0);	/* Unsupported output data type */
    
    /* Write the data to the audio device */
    Nb = write (audio_fd, (void *) Fbuf, Nbytes);
    if (Nb < Nbytes)
      UTerror ("%s: %s", PGM, ADM_WriteErr);
  }

  /* One last write of zero length (increments EOF counter) */
  write (audio_fd, (void *) Fbuf, 0);

  /* Wait for the output to finish */
  ADdrain (audio_fd);

  /* Report overloads */
  if (Novld > 0)
    UTwarn (ADMF_NOvld, PGM, Novld);

  /* Deallocate the buffer */
  UTfree (Fbuf);

  return;
}
