/*-------------- Telecommunications & Signal Processing Lab ---------------
                             McGill University

Routine:
  AFILE *AFsetRead (FILE *fp, int Ftype, struct AF_read *AFr, int Fix)

Purpose:
  Set up the parameters in an audio file structure for reading

Description:
  This routine checks the input parameters and puts them into a new audio file
  structure.  The data length (Ldata) and number of samples (Nsamp) parameters
  are checked for consistency and modified if necessary.

  The data length is the size of the area (in bytes) in the file reserved for
  for storing the data.  The data need not fully occupy this area.  If Ldata is
  undefined (equal to AF_LDATA_UNDEF), it is set to the size of the file, less
  the size of the header.  Ldata is checked against the actual size of the
  file.  It is an error for this value to exceed the actual space in the file.

  The number of samples Nsamp is used to calculate the actual amount of space
  space occupied by the data (Nsamp is multiplied by the data element size).
  If this value exceeds the space available, this is an error.  If Nsamp is
  undefined (equal to AF_NSAMP_UNDEF), it is determined from Ldata, setting it
  equal to the integer number of samples that fit into Ldata.  If Nsamp is so
  determined, a warning message is issued if Ldata is not a multiple of the
  data element size.  A warning message is also printed if Nsamp is not a
  multiple of the number of channels.

  For certain file types, the Nsamp value needs to be "fixed".  Optional fixes
  can be specified for the case that Nsamp is larger or smaller than the data
  space available in the file.  The latter case is not normally an error.  If
  a fixup is done, a warning message is printed.  A fixup can also be specified
  for values of Nsamp which are too large.

  If the data does not have fixed length elements (for instance, text data),
  Nsamp must be specified.  In that case, Ldata has no specific meaning.  If
  defined, Ldata is checked against the file size.

  The file must be positioned at the start of the audio data.

Parameters:
  <-  AFILE *AFp
      Audio file pointer for the audio file.  This routine allocates the space
      for this structure.  If an error is detected, a NULL pointer is returned.
   -> int Ftype
      File type: FT_NH, FT_AU, FT_WAVE, etc.
  <-> struct AF_read *AFr
      Structure with data and file parameters
   -> int Fix
      Fix flag, sum of the following subcodes
         AF_FIX_NSAMP_HIGH - Reduce Nsamp if it larger than the data
                             space available
         AF_FIX_NSAMP_LOW  - Increase Nsamp if it is smaller than the
                             data space available
         AF_FIX_LDATA_HIGH - Reduce Ldata if it is larger than the data
                           space available

Author / revision:
  P. Kabal  Copyright (C) 1999
  $Revision: 1.30 $  $Date: 1999/06/07 16:29:12 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFsetRead.c 1.30 1999/06/07 AFsp-v5r0 $";

#include <assert.h>
#include <math.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/nucleus.h>
#include <libtsp/AFdataio.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#define AF_DATA_LENGTHS
#include <libtsp/AFpar.h>

static long int
AF_Nrec (FILE *fp);
static int
AF_checkLdata (FILE *fp, int Lw, long int *Dstart, struct AF_ndata *NData,
	       int Fix);
static double
AF_srateInfo (const struct AF_info *Hinfo, double Sfreq);
static void
AF_setInfo (const struct AF_info *Hinfo, struct AF_info *AFinfo);
int
AF_setLdata (FILE *fp, long int Dstart, struct AF_ndata *NData, int Fix);
int
AF_setNsamp (FILE *fp, int Lw, struct AF_ndata *NData, int Fix);
static int
AF_checkNsamp (int Lw, struct AF_ndata *NData, int Fix);


AFILE *
AFsetRead (FILE *fp, int Ftype, struct AF_read *AFr, int Fix)

{
  AFILE *AFp;
  long int Dstart;
  int Lw;

  assert (Ftype > 0 && Ftype < NFT);
  assert (AFr->DFormat.Format > 0 && AFr->DFormat.Format < NFD);

/* Check data consistency */
  Lw = AF_DL[AFr->DFormat.Format];
  if (AF_checkLdata (fp, Lw, &Dstart, &AFr->NData, Fix))
    return NULL;

/* Sampling frequency */
  AFr->Sfreq = AF_srateInfo (&AFr->Hinfo, AFr->Sfreq);

/* Set the parameters for file access */
  AFp = (AFILE *) UTmalloc (sizeof (AFILE));
  AFp->fp = fp;
  AFp->Op = FO_RO;
  AFp->Error = AF_NOERR;
  AFp->Ftype = Ftype;
  AFp->Format = AFr->DFormat.Format;
  if (Lw <= 1)
    AFp->Swapb = DS_NATIVE;
  else
    AFp->Swapb = UTswapCode (AFr->DFormat.Swapb);
  AFp->Sfreq = AFr->Sfreq;
  AFp->ScaleF = AFr->DFormat.ScaleF;
  AFp->Nchan = AFr->NData.Nchan;
  AFp->Start = Dstart;
  AFp->Isamp = 0L;
  AFp->Nsamp = AFr->NData.Nsamp;
  AFp->Novld = 0L;
  AF_setInfo (&AFr->Hinfo, &AFp->Hinfo);

  return AFp;
}

/* Check file data size and number of samples */


static int
AF_checkLdata (FILE *fp, int Lw, long int *Dstart, struct AF_ndata *NData,
	       int Fix)

{
  int ErrCode;

/* Preliminary checks */
  if (NData->Ldata != AF_LDATA_UNDEF && NData->Ldata < 0) {
    UTwarn ("AFsetRead - %s: \"%ld\"", AFM_BadDLen, NData->Ldata);
    return 1;
  }
  if (NData->Nsamp != AF_NSAMP_UNDEF && NData->Nsamp < 0) {
    UTwarn ("AFsetRead - %s: \"%ld\"", AFM_BadNSamp, NData->Nsamp);
    return 1;
  }

/* Save the current (Start-of-Data) position */
  *Dstart = 0L;
  if (FLseekable (fp)) {
    ErrCode = 0;
    *Dstart = AFtell (fp, &ErrCode);
    if (ErrCode)
      return 1;

    /* Find Ldata and Dstart */
    if (Lw != 0) {
      /* Binary, seekable file */
      if (AF_setLdata (fp, *Dstart, NData, Fix))
	return 1;
    }
  }

/* Set Nsamp if necessary */
  if (AF_setNsamp (fp, Lw, NData, Fix))
    return 1;

/* Samples / channels consistency check */
  if (NData->Nchan <= 0L) {
    UTwarn ("AFsetRead - %s: \"%ld\"", AFM_BadNChan, NData->Nchan);
    return 1;
  }
  if (NData->Nsamp != AF_NSAMP_UNDEF && NData->Nsamp % NData->Nchan != 0) {
    UTwarn ("AFsetRead - %s", AFM_NSampNChan);
    UTwarn (AFMF_NSampNChan, "           ", NData->Nsamp, NData->Nchan);
  }

  return 0;
}

/* Set Ldata for binary, seekable files */


int
AF_setLdata (FILE *fp, long int Dstart, struct AF_ndata *NData, int Fix)

{
  long int Dspace;

  Dspace = FLfileSize (fp) - Dstart;
  if (Dspace < 0L)
    return 1;

  /* Set Ldata if necessary */
  if (NData->Ldata == AF_LDATA_UNDEF)
    NData->Ldata = Dspace;
  else if (NData->Ldata > Dspace) {
    if (Fix & AF_FIX_LDATA_HIGH) {
      UTwarn ("AFsetRead - %s", AFM_FixDLen);
      NData->Ldata = Dspace;
    }
    else {
      UTwarn ("AFsetRead - %s", AFM_ErrDLen);
      return 1;
    }
  }

  return 0;
}

/* Set and check Nsamp */

int
AF_setNsamp (FILE *fp, int Lw, struct AF_ndata *NData, int Fix)

{
/* Set Nsamp if necessary */
  if (Lw != 0) {
    if (NData->Nsamp == AF_NSAMP_UNDEF && NData->Ldata != AF_LDATA_UNDEF) {
      NData->Nsamp = NData->Ldata / Lw;
      if (Lw * NData->Nsamp != NData->Ldata)
	UTwarn ("AFsetRead - %s", AFM_NonIntNSamp);
    }
    else if (NData->Nsamp != AF_NSAMP_UNDEF) {

      /* Check Nsamp against Ldata */
      if (AF_checkNsamp (Lw, NData, Fix))
	return 1;
    }

  }
  else if (NData->Nsamp == AF_NSAMP_UNDEF && FLseekable (fp)) {
    NData->Nsamp = AF_Nrec (fp);
    if (NData->Nsamp == AF_NSAMP_UNDEF)
      return 1;	/* Error in AF_Nrec */
  }

  /* Check if an "unknown" Nsamp is allowed */
  if (NData->Nsamp == AF_NSAMP_UNDEF && ! AFopt_NsampND) {
    UTwarn ("AFsetRead - %s", AFM_NoNSamp);
    return 1;
  }

  return 0;
}

/*
  This routine finds the number of text records in a file, starting at the
  current position in the file.  The file position is restored to the current
  position.
*/


static long int
AF_Nrec (FILE *fp)

{
  long int pos, Nrec;
  char *p;
  int ErrCode;

  ErrCode = 0;

/* Save the file position */
  pos = AFtell (fp, &ErrCode);

/* Read until End-of-file */
  for (Nrec = 0L; ; ++Nrec) {

    /* Read a line or partial line */
    p = AFgetLine (fp, &ErrCode);
    if (p == NULL || ErrCode)
      break;
  }

/* Reposition the file */
  if (AFseek (fp, pos, &ErrCode))
    return AF_NSAMP_UNDEF;

  return Nrec;
}

/* Check Nsamp against Ldata and fix if so desired */


static int
AF_checkNsamp (int Lw, struct AF_ndata *NData, int Fix)

{
  if (NData->Nsamp * Lw > NData->Ldata) {
    if (Fix & AF_FIX_NSAMP_HIGH) {
      UTwarn ("AFsetRead - %s", AFM_FixDLen);
      NData->Nsamp = NData->Ldata / Lw;
    }
    else {
      UTwarn ("AFsetRead - %s", AFM_ErrDLen);
      return 1;
    }
  }
  else if (NData->Nsamp * Lw < NData->Ldata && (Fix & AF_FIX_NSAMP_LOW)) {
    UTwarn ("AFsetRead - %s", AFM_LowDLen);
    NData->Nsamp = NData->Ldata / Lw;
  }

  return 0;
}

/* Scan the header information string for a sampling frequency */


#define ABSV(x)		(((x) < 0) ? -(x) : (x))


static double
AF_srateInfo (const struct AF_info *Hinfo, double Sfreq)

{
  const char *Csf;
  double Fv;

  /* Check "sample_rate:", if the sampling frequency is integer-valued */
  if (Sfreq == floor(Sfreq)) {

    /* Scan the information string for a "sample_rate:" record */
    Csf = AFgetHrec ("sample_rate:", Hinfo);
    if (Csf != NULL) {
      if (STdec1double (Csf, &Fv))
	UTwarn ("AFsetRead - %s", AFM_BadSRate);
      else if (ABSV(Fv - Sfreq) <= 0.5)
	Sfreq = Fv;
      else
	UTwarn ("AFsetRead - %s, %g : %g", AFM_MisSRate, Sfreq, Fv);
    }
  }

  if (Sfreq <= 0.0) {
    UTwarn (AFMF_BadSFreq, "AFsetRead -", Sfreq, AF_SFREQ_DEFAULT);
    Sfreq = AF_SFREQ_DEFAULT;
  }

  return Sfreq;
}

/* Fill in the AFsp Information structure */


static void
AF_setInfo (const struct AF_info *Hinfo, struct AF_info *AFinfo)

{
  int N;

  if (Hinfo == NULL || Hinfo->N <= 0) {
    AFinfo->Info = NULL;
    AFinfo->N = 0;
    AFinfo->Nmax = 0;
  }
  else {
    N = Hinfo->N;
    if (Hinfo->Info[N-1] != '\0')
      ++N;
    AFinfo->Info = (char *) UTmalloc (N);
    memcpy (AFinfo->Info, Hinfo->Info, Hinfo->N);
    AFinfo->Info[N-1] = '\0';	/* Make sure there is a terminating null */
    AFinfo->N = N;
    AFinfo->Nmax = N;
  }

  return;
}
